import matplotlib
import matplotlib.pyplot as plt
from matplotlib.figure import Figure
from mpl_toolkits.mplot3d import Axes3D
from mpl_toolkits.mplot3d.art3d import Poly3DCollection
from mpl_toolkits.mplot3d.art3d import Path3DCollection
from matplotlib.backends.backend_tkagg import FigureCanvasTkAgg
import numpy as np
from wizzi_utils import misc_tools


def get_RGB_color(color_str: str) -> tuple:
    """
    :param color_str: e.g. "red", "blue" ...
    :return: RGB color
    rgb_r = get_RGB_color(color)
    rgba_r = get_RGBA_color(color, opacity=1.0)
    bgr_r = get_BGR_color(color)
    print('{}: RGB={}, RGBA={}, BGR={}'.format(color, rgb_r, rgba_r, bgr_r))
    # orange: RGB=(255, 165, 0), RGBA=(1.0, 0.6470588235294118, 0.0, 1.0), BGR=(0, 165, 255)
    """
    rgb = RGBA_to_RGB(color_str)
    return rgb


def get_RGBA_color(color_str: str, opacity: float = 1.0) -> tuple:
    """
    :param color_str: e.g. "red", "blue" ...
    :param opacity: value from 0 to 1
    :return: rgba color
    rgb_r = get_RGB_color(color)
    rgba_r = get_RGBA_color(color, opacity=1.0)
    bgr_r = get_BGR_color(color)
    print('{}: RGB={}, RGBA={}, BGR={}'.format(color, rgb_r, rgba_r, bgr_r))
    # orange: RGB=(255, 165, 0), RGBA=(1.0, 0.6470588235294118, 0.0, 1.0), BGR=(0, 165, 255)
    """
    rgba = matplotlib.colors.to_rgba(color_str, alpha=opacity)
    return rgba


def get_BGR_color(color_str: str) -> tuple:
    """
    :param color_str: e.g. "red", "blue" ...
    :return: bgr color
    rgb_r = get_RGB_color(color)
    rgba_r = get_RGBA_color(color, opacity=1.0)
    bgr_r = get_BGR_color(color)
    print('{}: RGB={}, RGBA={}, BGR={}'.format(color, rgb_r, rgba_r, bgr_r))
    # orange: RGB=(255, 165, 0), RGBA=(1.0, 0.6470588235294118, 0.0, 1.0), BGR=(0, 165, 255)
    """
    bgr = RGBA_to_BGR(color_str)
    return bgr


def RGBA_to_RGB(rgba: [str, tuple]) -> tuple:
    """
    :param rgba: e.g. "orange" or (1.0, 0.6470588235294118, 0.0, 1.0)
    :return: rgb color format
    e.g. orange: RGBA=(1.0, 0.6470588235294118, 0.0, 1.0), RGB=(255, 165, 0)
    """
    rgb_normed = matplotlib.colors.to_rgb(rgba)
    rgb = tuple([int(round(255 * x)) for x in rgb_normed])
    return rgb


def RGBA_to_BGR(rgba: [tuple, str]) -> tuple:
    """
    :param rgba: e.g. "orange" or (1.0, 0.6470588235294118, 0.0, 1.0)
    :return: bgr color format
    e.g. orange RGBA=(1.0, 0.6470588235294118, 0.0, 1.0), BGR=(0, 165, 255)
    """
    rgb = RGBA_to_RGB(rgba)
    bgr = RGB_to_BGR(rgb)
    return bgr


def BGR_to_RGB(bgr: tuple) -> tuple:
    """
    :param bgr:
    :return: rgb
    RGB=(255, 165, 0), BGR=(0, 165, 255)
    """
    rgb = misc_tools.reverse_tuple_or_list(bgr)  # reverse the tuple order
    return rgb


def BGR_to_RGBA(bgr: tuple, opacity: float = 1.0) -> tuple:
    """
    :param bgr:
    :param opacity:
    :return: rgba
    BGR=(0, 165, 255), RGBA=(1.0, 0.6470588235294118, 0.0, 1.0)
    """
    rgb = BGR_to_RGB(bgr)
    rgba = RGB_to_RGBA(rgb, opacity)
    return rgba


def RGB_to_BGR(rgb: tuple) -> tuple:
    """
    :param rgb:
    :return: bgr
    BGR=(0, 165, 255), RGB=(255, 165, 0)
    """
    bgr = misc_tools.reverse_tuple_or_list(rgb)  # reverse the tuple order
    return bgr


def RGB_to_RGBA(rgb: tuple, opacity: float = 1.0) -> tuple:
    """
    :param rgb:
    :param opacity:
    :return: rgba
    RGB=(255, 165, 0), RGBA=(1.0, 0.6470588235294118, 0.0, 1.0)
    """
    rgba = tuple([x / 255 for x in rgb])
    rgba += (opacity,)
    return rgba


def get_x_ticks_list(x_low, x_high, p=10):
    ten_percent_jump = (x_high - x_low) / p
    x_ticks = [x_low + i * ten_percent_jump for i in range(p + 1)]
    return x_ticks


def get_random_color_map(n: int, opacity: float = 1.0) -> np.array:
    """
    get colors list uniform distribution
    :param n: how many colors
    :param opacity:
    :return: np array of size(n,4). each row in RGBA format
    e.g.
    print(misc_tools.to_str(get_random_color_map(n=3), 'random_color_map'))
    """
    colors_map = misc_tools.np_uniform(shape=(n, 4), lows=0, highs=1)
    colors_map[:, -1] = opacity
    return colors_map


def screen_dims():
    """
    should pip install PIL
    :return:
    """
    from PIL import ImageGrab
    try:
        img = ImageGrab.grab()
        window_w, window_h = img.size
    except (ValueError, Exception):
        window_w, window_h = 0, 0
        print('screen_dims()::Failed getting screen_dims')
    return window_w, window_h


def move_figure(fig, x: int, y: int):
    """Move figure's upper left corner to pixel (x, y)"""
    try:
        x, y = int(x), int(y)
        new_geom = "+{}+{}".format(x, y)
        backend = matplotlib.get_backend()
        if backend == 'TkAgg':
            manager = fig.canvas.manager
            manager.window.wm_geometry(new_geom)
        elif backend == 'WXAgg':
            fig.canvas.manager.window.SetPosition((x, y))
        else:
            # This works for QT and GTK
            # You can also use window.setGeometry
            fig.canvas.manager.window.move(x, y)
    except (ValueError, Exception):
        print('move_figure::Failed Moving figure to ({},{})'.format(x, y))
        return
    return


def move_plot(fig, where: str = 'top_left'):
    """
    :param fig:
    :param where:
        top_right, top_center, top_left, bottom_right, bottom_center, bottom_left
    :return:
    """
    try:
        window_w, window_h = screen_dims()  # screen dims in pixels
        fig_w, fig_h = fig.get_size_inches() * fig.dpi  # fig dims in pixels
        task_bar_offset = 100  # about 100 pixels due to task bar

        x, y = 0, 0  # top_left: default

        if where == 'top_center':
            x, y = (window_w - fig_w) / 2, 0
        elif where == 'top_right':
            x, y = (window_w - fig_w), 0
        elif where == 'bottom_left':
            x, y = 0, window_h - fig_h - task_bar_offset
        elif where == 'bottom_center':
            x, y = (window_w - fig_w) / 2, window_h - fig_h - task_bar_offset
        elif where == 'bottom_right':
            x, y = (window_w - fig_w), window_h - fig_h - task_bar_offset
        move_figure(fig=fig, x=x, y=y)
    except (ValueError, Exception):
        print('move_plot::Failed Moving figure to {}'.format(where))
        return
    return


# 2d plots
def plot_2d_iterative_figure(
        rows: int = 1,
        cols: int = 1,
        main_title: str = None,
        sub_titles: list = None,
        labels: list = None,
        default_color: str = 'green',
        resize: float = 0,
        plot_location: str = None,
        x_y_lims: list = None,
        add_center: dict = False,
        zoomed: bool = False,
        render_d: dict = None
) -> (matplotlib.figure, list, list):
    """
    THIS is for building a figure with 1 or more subplots that changes each iteration.
        this function just builds the frame. use update_subplots to insert\change the data of the scatters
    :param rows:
    :param cols:
    :param main_title:
    :param sub_titles: if not None, sub title to each sub plots. |sub_titles| must be equal to rows*cols
    :param labels: if labels are given to each subplot, legend will be on
    :param default_color: when updating, color is not mandatory. if non given, default color will remain
    :param resize: if not 0, figure size *= resize
    :param plot_location: top_right, top_center, top_left, bottom_right, bottom_center, bottom_left or None
    :param x_y_lims: limits of x and y axes. list of 4 ints: x_left, x_right, y_bottom, y_top
    :param add_center: dict. if not None - add center with the params in the dict
        e.g. add_center = {'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'}
    :param zoomed: bool - full screen or not
    :param render_d: dict - if not None: show plot
        block mandatory. pause optional.
        e.g. render = {'block': False, 'pause': 0.0001}
    :return:
    figure
    list of axes
    list of scatters (1 for each ax in axes - this will be used to update the data)

    see full example in update_2d_scatters comments
    """
    plt.close('all')
    figsize = (6.4, 4.8) if resize == 0 else (6.4 * resize, 4.8 * resize)  # default figsize=(6.4, 4.8)
    fig, axes = plt.subplots(nrows=rows, ncols=cols, sharex=False, sharey=False, figsize=figsize)
    axes_list = [axes] if (rows == 1 and cols == 1) else axes.flatten().tolist()

    if sub_titles is not None:
        err_msg = '|sub_titles| must be equal to rows*cols ({}!={})'.format(len(sub_titles), rows * cols)
        assert len(sub_titles) == len(axes_list), err_msg
    if labels is not None:
        err_msg = '|labels| must be equal to rows*cols ({}!={})'.format(len(labels), rows * cols)
        assert len(labels) == len(axes_list), err_msg

    if main_title is not None:
        fig.suptitle(main_title)
    if plot_location is not None:
        move_plot(fig, where=plot_location)

    scatters = []
    for i, ax in enumerate(axes_list):
        if sub_titles is not None:
            ax.set_title(sub_titles[i])
        ax.set_aspect('equal', adjustable='box')

        if x_y_lims is not None:
            ax.set_xlim(left=x_y_lims[0], right=x_y_lims[1])
            ax.set_ylim(bottom=x_y_lims[2], top=x_y_lims[3])
            # scatters.append(sc)

        x_left, x_right = ax.get_xlim()
        y_bottom, y_top = ax.get_ylim()

        cx = int((x_right - x_left) / 2) + x_left
        cy = int((y_top - y_bottom) / 2) + y_bottom

        ax.set_xticks([x_left, cx, x_right])
        ax.set_yticks([y_bottom, cy, y_top])

        if add_center is not None:
            ax.scatter(
                [cx], [cy],
                c=add_center['c'], marker=add_center['marker'],
                s=add_center['marker_size'], label='{}({},{})'.format(add_center['label'], cx, cy)
            )

        # init data scatter for this axis
        if labels is not None:
            sc = ax.scatter([10], [10], c=default_color, marker='.', label=labels[i])
            ax.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5, edgecolor='black')
        else:
            sc = ax.scatter([10], [10], c=default_color, marker='.')
        scatters.append(sc)

    if zoomed:
        wm = plt.get_current_fig_manager()
        wm.window.state('zoomed')

    if render_d is not None:
        if 'pause' in render_d:
            render(block=render_d['block'], pause=render_d['pause'])
        else:
            render(block=render_d['block'])
    return fig, axes_list, scatters


def update_2d_scatters(
        scatters: list,
        datum: list,
        colors_sets: list = None,
        new_title: str = None,
        save_img_path: str = None,
        render_d: dict = None
) -> None:
    """
    :param scatters: list of scatters to update
    :param datum: list of data per scatter.
            |datum| == |scatters|
            data could be empty list or None
    :param colors_sets: list of colors per scatter.
            if colors_sets is not None: |colors_sets| == |scatters| == |datum|
            color_set could be:
                empty list or None - former color will remain (first is default color)
                list of size 1 - RGBA color. all points will get this color
                list of size |data| - RGBA colors. each point will get it's corresponding color
    :param new_title: changed main title
    :param save_img_path: if not none, save the fig to this path
    :param render_d: dict - if not None: show plot
        block mandatory. pause optional.
        e.g. render = {'block': False, 'pause': 0.0001}
    :return:
    example
    fig, axes_list, scatters = plot_2d_iterative_figure(
        rows=1,
        cols=2,
        main_title='1x1 scatter',
        sub_titles=['left cam', 'right cam'],
        labels=['MVS1', 'MVS2'],
        default_color='blue',
        resize=0,
        plot_location='top_center',
        x_y_lims=[0, 640, 0, 480],  # fits cv img
        add_center={'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'},
        zoomed=False,
        render_d={'block': False, 'pause': 0.0001}
    )
    iters = 10
    for i in range(iters):
        misc_tools.sleep(seconds=1)
        block = (i == iters - 1)  # last iter
        # emulate data of each scatter - say number of points is 3
        datum, colors_sets = [], []
        for j in range(len(scatters)):
            data_j = misc_tools.np_random_integers(low=0, high=480, size=(3, 2))
            if j == 0:  # cam0
                colors = [get_RGBA_color('y')]  # 1 color for all
            else:  # cam1
                colors = get_random_color_map(3)  # random color for each point
                # # fixed color per point
                # colors = [get_RGBA_color('r'), get_RGBA_color('b'), get_RGBA_color('g')]
            if i == 2 and j == 1:  # emulate in iter 2, cam 1 found no data
                data_j = []
                colors = []
            datum.append(data_j)
            colors_sets.append(colors)

        update_2d_scatters(
            scatters=scatters,
            datum=datum,
            colors_sets=colors_sets,
            new_title='iter {}'.format(i),
            render_d={'block': block, 'pause': 0.0001}
        )
    """
    assert len(datum) == len(scatters), 'data per scatter is required.'
    err_msg = 'colors_sets should be none or same size as scatters and datum'
    assert colors_sets is None or len(colors_sets) == len(scatters) == len(datum), err_msg

    if new_title is not None:
        plt.suptitle(new_title)
    for i in range(len(scatters)):
        sc_i = scatters[i]
        if datum[i] is not None and len(datum[i]) > 0:
            sc_i._offsets = datum[i]
            if colors_sets is not None and colors_sets[i] is not None and len(colors_sets[i]) > 0:
                sc_i._facecolors = colors_sets[i]
                sc_i._edgecolors = colors_sets[i]

    if save_img_path is not None:
        plt.savefig(save_img_path, dpi=200, bbox_inches='tight')
        print('saved to {}.png'.format(save_img_path))

    if render_d is not None:
        if 'pause' in render_d:
            render(block=render_d['block'], pause=render_d['pause'])
        else:
            render(block=render_d['block'])
    return


def plot_2d_scatter(
        data: np.array,
        colors: list = None,
        data_label: str = None,
        main_title: str = None,
        sub_title: str = None,
        def_color: str = None,
        resize: float = 0,
        plot_location=None,
        x_y_lims=None,
        save_img_path: str = None,
        add_center: dict = False,
        zoomed: bool = False,
) -> None:
    """
    see documentation in plot_2d_iterative_figure() and update_2d_scatters()

    example:
    data_j = misc_tools.np_random_integers(low=0, high=480, size=(3, 2))
    colors = [get_RGBA_color('r'), get_RGBA_color('b'), get_RGBA_color('g')]
    plot_2d_scatter(
        data=data_j,
        colors=colors,
        data_label='SCP',
        main_title='1x1 scatter',
        sub_title='normal 2d scatter',
        def_color='green',
        resize=0,
        plot_location='top_center',
        x_y_lims=[0, 640, 0, 480],  # fits cv img,
        save_img_path='./test',
        add_center={'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'},
        zoomed=False,
    )
    """
    fig, axes_list, scatters = plot_2d_iterative_figure(
        rows=1,
        cols=1,
        main_title=main_title,
        sub_titles=[sub_title],
        labels=[data_label],
        default_color=def_color,
        resize=resize,
        plot_location=plot_location,
        x_y_lims=x_y_lims,
        add_center=add_center,
        zoomed=zoomed,
        render_d=None
    )
    update_2d_scatters(
        scatters=scatters,
        datum=[data],
        colors_sets=[colors],
        new_title=None,
        save_img_path=save_img_path,
        render_d={'block': True}
    )
    return


def plot_x_y_std(data_x: np.array, groups: list, title: str = None, x_label: str = 'Size', y_label: str = 'Error',
                 save_path: str = None, show_plot: bool = True, with_shift: bool = False):
    """
    data_x: x values
    groups: list of groups s.t. each tuple(y values, y std, color, title)  y std could be None
    example:
        data_x = [10, 20, 30]
        C_errors = [5, 7, 1]
        C_errors_stds = [2, 1, 0.5]
        group_c = (C_errors, C_errors_stds, 'g', 'C')
        U_errors = [10, 8, 3]
        U_errors_vars = [4, 3, 1.5]
        group_u = (U_errors, U_errors_vars, 'r', 'U')
        groups = [group_c, group_u]
        title = 'bla'
        plot_x_y_std(data_x, groups, title)
    :return:
    """
    data_x_last = data_x  # in order to see all STDs, move a little on the x axis
    data_x_jump = 0.5
    data_x_offset = - int(len(groups) / 2) * data_x_jump
    line_style = {"linestyle": "-", "linewidth": 1, "markeredgewidth": 2, "elinewidth": 1, "capsize": 4}
    for i, group in enumerate(groups):
        data_y, std_y = group[0], group[1]  # std_y could be None
        color, label = group[2], group[3]
        if with_shift:  # move x data for each set a bit so you can see it clearly
            dx_shift = [x + i * data_x_jump + data_x_offset for x in data_x]
            data_x_last = dx_shift
        plt.errorbar(data_x_last, data_y, std_y, color=color, fmt='.', label=label, **line_style)

    plt.grid()
    plt.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5)
    if title is not None:
        plt.title(title)
    plt.xticks(data_x)
    plt.xlabel(x_label)
    plt.ylabel(y_label)
    if save_path is not None:
        plt.savefig(save_path, dpi=100, bbox_inches='tight')
        print('\tsaved to {}.png'.format(save_path))
    plt.pause(0.0001)
    if show_plot:
        plt.show(block=True)
    plt.cla()
    return


def histogram(values: np.array, title: str, save_path: str = None, bins_n: int = 50):
    """ plots a histogram """
    plt.hist(values, bins_n, density=False, facecolor='blue', alpha=0.75)
    plt.xlabel('Values')
    plt.ylabel('Bin Count')
    plt.title(title)
    plt.grid(True)
    if save_path is not None:
        plt.savefig(save_path, dpi=100, bbox_inches='tight')
        print('\tsaved to {}.png'.format(save_path))
    plt.show()
    plt.cla()
    return


def compare_images_sets(set_a, set_b, title: str = None):
    """
    build for images BEFORE transform:
    notice images should be in the format:
        gray scale mnist: [number of images, 28, 28]
        RGB  Cifar10    : [number of images, 32, 32, 3]

    :param set_a: array (nd\torch) of images
    :param set_b: array (nd\torch) of images
    :param title: plot title
    plot set a of images in row 1 and set b in row 2
    set_a and set_b can be ndarray or torch arrays
    example:
        from torchvision import datasets
        # choose data set - both work
        # data_root = path to the data else download
        data_root = '../../2019SGD/Datasets/'
        # dataset = datasets.MNIST(root=data_root, train=False, download=False)
        dataset = datasets.CIFAR10(root=data_root, train=False, download=False)
        set_a = dataset.data[:3]
        set_b = dataset.data[10:50]
        compare_images_sets(set_a, set_b)
        set_a = dataset.data[0:3]
        set_b = dataset.data[0:3]
        compare_images_sets(set_a, set_b)
    """
    n_cols = max(set_a.shape[0], set_b.shape[0])
    fig, axes = plt.subplots(nrows=2, ncols=n_cols, sharex='all', sharey='all', figsize=(15, 4))
    for images, row in zip([set_a, set_b], axes):
        for img, ax in zip(images, row):
            ax.imshow(np.squeeze(img), cmap='gray')
            ax.get_xaxis().set_visible(False)
            ax.get_yaxis().set_visible(False)
    if title is not None:
        plt.title(title)
    plt.show()
    return


def compare_images_multi_sets_squeezed(sets_dict: dict, title: str = None) -> str:
    """
    build for images AFTER transform:
    notice images should be in the format:
        gray scale mnist: [number of images, 1, 28, 28]
        RGB  Cifar10    : [number of images, 3, 32, 32]

    :param sets_dict: each entry in dict is title, set of images(np/tensor)
    :param title: for plot
    :return str with details which set in each row
    plot sets of images in rows
    example:
        import torch
        from torchvision import datasets
        import torchvision.transforms as transforms
        transform = transforms.Compose([transforms.ToTensor(), ])
        # choose data set - both work
        # data_root = path to the data else download
        data_root = '../../2019SGD/Datasets/'
        # dataset = datasets.MNIST(root=data_root, train=False, download=False, transform=transform)
        dataset = datasets.CIFAR10(root=data_root, train=False, download=False, transform=transform)
        data_loader = torch.utils.data.DataLoader(dataset, batch_size=32, shuffle=True, num_workers=2)
        images32, labels = iter(data_loader).next()

        images = images32[:16]  # imagine the first 16 are base images and predicted_images are the model output
        predicted_images = images32[16:32]
        d = {'original_data': images, 'predicted_data': predicted_images}
        print(compare_images_multi_sets_squeezed(d))
    """
    from wizzi_utils import torch_tools as tt
    from torchvision.utils import make_grid
    import torch
    for k, v in sets_dict.items():
        if isinstance(sets_dict[k], np.ndarray):
            sets_dict[k] = tt.numpy_to_torch(sets_dict[k])

    all_sets = None
    msg = ''
    set_len = 0
    msg_base = 'row {}: {}, '

    for i, (k, v) in enumerate(sets_dict.items()):
        all_sets = v if all_sets is None else torch.cat((all_sets, v), 0)
        msg += msg_base.format(i, k)
        set_len = v.shape[0]

    grid_images = make_grid(all_sets, nrow=set_len)
    if title is not None:
        plt.title(title)
    plt.axis('off')
    plt.imshow(np.transpose(tt.torch_to_numpy(grid_images), (1, 2, 0)))
    plt.show()
    return msg


def render(block: bool = False, pause: float = 0.0001):
    plt.draw()
    plt.show(block=block)
    plt.pause(pause)
    return


# 3d plots
def plot_3d_iterative_figure(
        scatter_dict: dict,
        main_title: str = None,
        resize: float = 0,
        plot_location: str = None,
        x_y_z_lims: list = None,
        fig_face_color: str = None,
        ax_background: str = None,
        ax_labels_and_ticks_c: str = None,
        add_center: dict = None,
        zoomed: bool = False,
        view: dict = None,
        render_d: dict = None
) -> (matplotlib.figure, Axes3D, Path3DCollection, FigureCanvasTkAgg):
    """
    THIS is for building a figure with 1 or more subplots that changes each iteration.
        this function just builds the frame. use update_subplots to insert\change the data of the scatters
    :param scatter_dict: dict with mandatory entries of the main scatter
        e.g. {'c': 'b', 'marker_size': 1, 'marker': 'o', 'label': 'MVS'} # c for color
    :param main_title:
    :param resize: if not 0, figure size *= resize
    :param plot_location: top_right, top_center, top_left, bottom_right, bottom_center, bottom_left or None
    :param x_y_z_lims: limits of x, y and z axes. list of 6 ints: x_left, x_right, y_bottom, y_top, z_in, z_out
    :param fig_face_color: color of the whole background - default is white
    :param ax_background: color of the axes background - default is white
    :param ax_labels_and_ticks_c: color of the ticks and axes labels
    :param add_center: dict. if not None - add center with the params in the dict
        e.g. add_center = {'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'}
    :param zoomed: bool - full screen or not
    :param view: dict - view onto the world. e.g. {'azim': 90.0, 'elev': -100.0}
    :param render_d: dict - if not None: show plot
        block mandatory. pause optional.
        e.g. render = {'block': False, 'pause': 0.0001}
    :return:
    figure
    axes3d
    Path3DCollection (pointer to the scatter data) - used for updating
    FigureCanvasTkAgg - used to change title
    see full example in update_3d_scatters comments
    """
    plt.close('all')
    figsize = (6.4, 4.8) if resize == 0 else (6.4 * resize, 4.8 * resize)  # default figsize=(6.4, 4.8)
    fig = plt.figure(figsize=figsize)
    fig_canvas = fig.canvas
    ax = Axes3D(fig)
    if fig_face_color is not None:
        ax.set_facecolor(fig_face_color)

    if ax_background is not None:
        ax_background = get_RGBA_color(ax_background)
        ax.w_xaxis.set_pane_color(ax_background)
        ax.w_yaxis.set_pane_color(ax_background)
        ax.w_zaxis.set_pane_color(ax_background)

    if main_title is not None:
        fig.suptitle(main_title)
    if plot_location is not None:
        move_plot(fig, where=plot_location)

    if x_y_z_lims is not None:
        ax.set_xlim3d(left=x_y_z_lims[0], right=x_y_z_lims[1])
        ax.set_ylim3d(bottom=x_y_z_lims[2], top=x_y_z_lims[3])
        ax.set_zlim3d(bottom=x_y_z_lims[4], top=x_y_z_lims[5])

    x_left, x_right = ax.get_xlim()
    y_bottom, y_top = ax.get_ylim()
    z_in, z_out = ax.get_zlim()

    cx = int((x_right - x_left) / 2) + x_left
    cy = int((y_top - y_bottom) / 2) + y_bottom
    cz = int((z_out - z_in) / 2) + z_in

    ax.set_xticks([x_left, cx, x_right])
    ax.set_yticks([y_bottom, cy, y_top])
    ax.set_zticks([z_in, cz, z_out])

    if ax_labels_and_ticks_c is not None:
        ax.w_xaxis.line.set_color(ax_labels_and_ticks_c)
        ax.w_yaxis.line.set_color(ax_labels_and_ticks_c)
        ax.w_zaxis.line.set_color(ax_labels_and_ticks_c)
        ax.tick_params(axis='x', colors=ax_labels_and_ticks_c)
        ax.tick_params(axis='y', colors=ax_labels_and_ticks_c)
        ax.tick_params(axis='z', colors=ax_labels_and_ticks_c)
        ax.set_xlabel("X", color=ax_labels_and_ticks_c)
        ax.set_ylabel("Y", color=ax_labels_and_ticks_c)
        ax.set_zlabel("Z", color=ax_labels_and_ticks_c)
    else:
        ax.set_xlabel("X")
        ax.set_ylabel("Y")
        ax.set_zlabel("Z")

    if add_center is not None:
        ax.scatter(
            cx, cy, cz,
            c=add_center['c'], marker=add_center['marker'],
            s=add_center['marker_size'], label='{}({},{},{})'.format(add_center['label'], cx, cy, cz)
        )

    iterative_scatter = ax.scatter(
        0, 0, 0,
        c=scatter_dict['c'], marker=scatter_dict['marker'],
        s=scatter_dict['marker_size'], label=scatter_dict['label']
    )

    plt.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5)

    if view is not None:
        ax.view_init(azim=view['azim'], elev=view['elev'])

    if zoomed:
        wm = plt.get_current_fig_manager()
        wm.window.state('zoomed')

    if render_d is not None:
        if 'pause' in render_d:
            render(block=render_d['block'], pause=render_d['pause'])
        else:
            render(block=render_d['block'])
    return fig, ax, iterative_scatter, fig_canvas


def update_3d_scatters(
        scatter: Path3DCollection,
        fig_canvas: FigureCanvasTkAgg,
        data: np.array,
        colors: [list, np.array] = None,
        new_title: str = None,
        save_img_path: str = None,
        render_d: dict = None
) -> None:
    """
    :param scatter: the data scatter obj
    :param fig_canvas: used to change title
    :param data: array of 3d data.
    :param colors: list of colors. None, 1 color or |data| colors. rgba format
    :param new_title: changed main title
    :param save_img_path: if not none, save the fig to this path
    :param render_d: dict - if not None: show plot
        block mandatory. pause optional.
        e.g. render = {'block': False, 'pause': 0.0001}
    :return:
    example

    fig, ax, iterative_scatter, fig_canvas = plot_3d_iterative_figure(
        scatter_dict={'c': 'b', 'marker_size': 100, 'marker': 'o', 'label': 'MVS'},
        main_title='3d scatter plot',
        resize=1.5,
        plot_location='top_center',
        x_y_z_lims=[-20, 20, -20, 20, -20, 20],
        fig_face_color=None,
        ax_background=None,
        ax_labels_and_ticks_c=None,
        add_center={'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'},
        zoomed=False,
        view=None,
        render_d=None
    )

    # # CUSTOM ADD ON 1 - update each round
    center_mass_x_y = {"x1": 0.05, "y1": 0.95, }
    center_mass_label_base = "(xyz)={}"
    center_mass_label = ax.text2D(
        x=center_mass_x_y['x1'],
        y=center_mass_x_y['y1'],
        s=center_mass_label_base.format(np.zeros(3)),
        transform=ax.transAxes,
        color='green'
    )

    # # CUSTOM ADD ON 2 - done once
    add_cube_around_origin(ax, edge_len=4, add_labels=False)

    # should call legend if you want custom addons in it
    # plt.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5)

    render(block=False)  # not necessary - only if you want to see the plot before first data comes

    num_points = 5
    iters = 10
    for i in range(10):
        misc_tools.sleep(seconds=1)
        data_j = misc_tools.np_random_integers(low=-15, high=15, size=(num_points, 3))
        colors = get_random_color_map(num_points)
        block = (i == iters - 1)  # last iter

        update_3d_scatters(
            scatter=iterative_scatter,
            fig_canvas=fig_canvas,
            data=data_j,
            colors=colors,
            new_title='iter {}'.format(i),
            render_d=None
        )

        # # CUSTOM ADD ON 1 - update each round
        center_mass_label.set_text(center_mass_label_base.format(np.round(np.mean(data_j, axis=0), 2)))

        render(block=block)
    """
    fig_canvas.set_window_title(new_title)
    scatter._offsets3d = data.T

    if colors is not None:
        scatter._facecolor3d = colors
        scatter._edgecolor3d = colors

    if save_img_path is not None:
        plt.savefig(save_img_path, dpi=200, bbox_inches='tight')
        print('saved to {}.png'.format(save_img_path))

    if render_d is not None:
        if 'pause' in render_d:
            render(block=render_d['block'], pause=render_d['pause'])
        else:
            render(block=render_d['block'])
    return


def plot_3d_scatter(
        scatter_dict: dict,
        data: np.array,
        colors: [list, np.array] = None,
        main_title: str = None,
        resize: float = 0,
        plot_location: str = None,
        x_y_z_lims: list = None,
        fig_face_color: str = None,
        ax_background: str = None,
        ax_labels_and_ticks_c: str = 'black',
        add_center: dict = False,
        zoomed: bool = False,
        view: dict = None,
        save_img_path: str = None,
) -> None:
    """
    see documentation in plot_3d_iterative_figure() and update_3d_scatters()

    example:
    num_points = 50
    data_j = misc_tools.np_random_integers(low=-15, high=15, size=(num_points, 3))
    plot_3d_scatter(
        scatter_dict={'c': 'b', 'marker_size': 100, 'marker': 'o', 'label': 'MVS'},
        data=data_j,
        colors=get_random_color_map(num_points),
        main_title='3d scatter plot',
        resize=1.5,
        plot_location='top_center',
        x_y_z_lims=[-20, 20, -20, 20, -20, 20],
        fig_face_color=None,
        ax_background=None,
        ax_labels_and_ticks_c=None,
        add_center={'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'},
        zoomed=False,
        view=None,
        save_img_path='./3dplot'
    )
    """
    fig, ax, iterative_scatter, fig_canvas = plot_3d_iterative_figure(
        scatter_dict=scatter_dict,
        main_title=main_title,
        resize=resize,
        plot_location=plot_location,
        x_y_z_lims=x_y_z_lims,
        fig_face_color=fig_face_color,
        ax_background=ax_background,
        ax_labels_and_ticks_c=ax_labels_and_ticks_c,
        add_center=add_center,
        zoomed=zoomed,
        view=view,
        render_d=None
    )

    update_3d_scatters(
        scatter=iterative_scatter,
        fig_canvas=fig_canvas,
        data=data,
        colors=colors,
        new_title=None,
        save_img_path=save_img_path,
        render_d={'block': True}
    )
    return


def plot_3d_cube(axes: Axes3D, cube_definition: list, color='b', label='cube', add_labels=False):
    """
    :param axes:
    :param cube_definition: list of np.arrays
        4 points of the cube (1 corner node and 3 nodes to the left right and top)
    :param color:
    :param label:
    :param add_labels: add cube corners label on figure
    :return:
    """

    points = []
    points += cube_definition
    vectors = [
        cube_definition[1] - cube_definition[0],
        cube_definition[2] - cube_definition[0],
        cube_definition[3] - cube_definition[0]
    ]

    points += [cube_definition[0] + vectors[0] + vectors[1]]
    points += [cube_definition[0] + vectors[0] + vectors[2]]
    points += [cube_definition[0] + vectors[1] + vectors[2]]
    points += [cube_definition[0] + vectors[0] + vectors[1] + vectors[2]]

    points = np.array(points)

    edges = [
        [points[0], points[3], points[5], points[1]],
        [points[1], points[5], points[7], points[4]],
        [points[4], points[2], points[6], points[7]],
        [points[2], points[6], points[3], points[0]],
        [points[0], points[2], points[4], points[1]],
        [points[3], points[6], points[7], points[5]]
    ]

    faces = Poly3DCollection(edges, linewidths=1, edgecolors='k')
    color_rgba = matplotlib.colors.to_rgba(color, alpha=0.1)
    faces.set_facecolor(color_rgba)
    axes.add_collection3d(faces)

    # Plot the points themselves to force the scaling of the axes
    axes.scatter(points[:, 0], points[:, 1], points[:, 2], s=0.1, color=color, label=label)

    if add_labels:
        for p in points:
            x, y, z = p
            text = '{},{},{}'.format(x, y, z)
            axes.text(x, y, z, text, zdir=(1, 1, 1))
    return


def add_cube3d_around_origin(axes: Axes3D, edge_len: int, add_labels: bool = False):
    """
    :param axes:
    :param edge_len: cube edge size
    :param add_labels: add cube labels on the scene
    :return:
    """
    half_edge = int(edge_len / 2)
    xyz_bot_left = np.array([-half_edge, -half_edge, -half_edge], dtype=float)
    xyz_top_left = np.copy(xyz_bot_left)
    xyz_bot_right = np.copy(xyz_bot_left)
    xyz_bot_left_depth = np.copy(xyz_bot_left)
    xyz_top_left[1] += edge_len  # add just y
    xyz_bot_right[0] += edge_len  # add just x
    xyz_bot_left_depth[2] += edge_len  # add just z

    cube_4_edges = [xyz_bot_left, xyz_top_left, xyz_bot_right, xyz_bot_left_depth]
    plot_3d_cube(
        axes,
        cube_4_edges,
        label='cube(edge={})'.format(edge_len),
        add_labels=add_labels
    )
    return


def main():
    fig, ax, iterative_scatter, fig_canvas = plot_3d_iterative_figure(
        scatter_dict={'c': 'b', 'marker_size': 100, 'marker': 'o', 'label': 'MVS'},
        main_title='3d scatter plot',
        resize=1.5,
        plot_location='top_center',
        x_y_z_lims=[-20, 20, -20, 20, -20, 20],
        fig_face_color=None,
        ax_background=None,
        ax_labels_and_ticks_c=None,
        add_center={'c': 'orange', 'marker': 'x', 'marker_size': 150, 'label': 'Scene Center'},
        zoomed=False,
        view=None,
        render_d=None
    )

    # # CUSTOM ADD ON 1 - update each round
    center_mass_x_y = {"x1": 0.05, "y1": 0.95, }
    center_mass_label_base = "(xyz)={}"
    center_mass_label = ax.text2D(
        x=center_mass_x_y['x1'],
        y=center_mass_x_y['y1'],
        s=center_mass_label_base.format(np.zeros(3)),
        transform=ax.transAxes,
        color='green'
    )

    # # CUSTOM ADD ON 2 - done once
    add_cube3d_around_origin(ax, edge_len=4, add_labels=False)

    # should call legend if you want custom addons in it
    # plt.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5)

    render(block=False)  # not necessary - only if you want to see the plot before first data comes

    num_points = 5
    iters = 10
    for i in range(10):
        misc_tools.sleep(seconds=1)
        data_j = misc_tools.np_random_integers(low=-15, high=15, size=(num_points, 3))
        colors = get_random_color_map(num_points)
        block = (i == iters - 1)  # last iter

        update_3d_scatters(
            scatter=iterative_scatter,
            fig_canvas=fig_canvas,
            data=data_j,
            colors=colors,
            new_title='iter {}'.format(i),
            render_d=None
        )

        # # CUSTOM ADD ON 1 - update each round
        center_mass_label.set_text(center_mass_label_base.format(np.round(np.mean(data_j, axis=0), 2)))

        render(block=block)
    return


if __name__ == '__main__':
    main()



import numpy as np


def find_centers(A: np.array, k: int = 1) -> np.array:
    """
    :param A: nx(d+1) data array. A=X|y
    :param k: how many centers
    :return: centers kxd
    e.g.
        from wizzi_utils import misc_tools as mt
        import matplotlib.pyplot as plt
        A = np.zeros((4, 2))  # A square with origin 0
        A[0] = [-1, -1]
        A[1] = [-1, 1]
        A[2] = [1, -1]
        A[3] = [1, 1]
        print(mt.to_str(A, title='A'))
        centers = find_centers(A, k=1)
        print(mt.to_str(centers, title='centers'))
        X, y = mt.de_augment_numpy(A)
        X_c, y_c = mt.de_augment_numpy(centers)
        plt.scatter(X, y, color='g', marker='.', label='A')
        plt.scatter(X_c, y_c, color='r', marker='.', label='k==1')
        plt.legend(loc='upper right', ncol=1, fancybox=True, framealpha=0.5, edgecolor='black')
        plt.show()
    """
    from sklearn.cluster import KMeans
    k_means_obj = KMeans(n_clusters=k)
    k_means_obj.fit(A)
    centers = k_means_obj.cluster_centers_
    return centers


def main():
    return


if __name__ == '__main__':
    main()



import os
import datetime
from timeit import default_timer as timer
from typing import Callable
import cProfile
import pstats
import io
import numpy as np
import random
import inspect
import sys
import time
import pickle
from itertools import combinations

LINES = '-' * 80
CONST_COLOR_MAP = {
    'ResetAll': "\033[0m",
    'Bold': "\033[1m",
    'Dim': "\033[2m",
    'Underlined ': "\033[4m",
    'Blink      ': "\033[5m",
    'Reverse    ': "\033[7m",
    'Hidden     ': "\033[8m",

    'ResetBold': "\033[21m",
    'ResetDim': "\033[22m",
    'ResetUnderlined': "\033[24m",
    'ResetBlink': "\033[25m",
    'ResetReverse': "\033[27m",
    'ResetHidden': "\033[28m",

    'Default': "\033[39m",
    'Black': "\033[97m",
    'Red': "\033[31m",
    'Green': "\033[32m",
    'Yellow': "\033[33m",
    'Blue': "\033[34m",
    'Magenta': "\033[35m",
    'Cyan': "\033[36m",
    'LightGray': "\033[37m",
    'DarkGray': "\033[90m",
    'LightRed': "\033[91m",
    'LightGreen': "\033[92m",
    'LightYellow': "\033[93m",
    'LightBlue': "\033[94m",
    'LightMagenta': "\033[95m",
    'LightCyan': "\033[96m",
    'White': "\033[30m",

    'BackgroundDefault': "\033[49m",
    'BackgroundBlack': "\033[107m",
    'BackgroundRed': "\033[41m",
    'BackgroundGreen': "\033[42m",
    'BackgroundYellow': "\033[43m",
    'BackgroundBlue': "\033[44m",
    'BackgroundMagenta': "\033[45m",
    'BackgroundCyan': "\033[46m",
    'BackgroundLightGray': "\033[47m",
    'BackgroundDarkGray': "\033[100m",
    'BackgroundLightRed': "\033[101m",
    'BackgroundLightGreen': "\033[102m",
    'BackgroundLightYellow': "\033[103m",
    'BackgroundLightBlue': "\033[104m",
    'BackgroundLightMagenta': "\033[105m",
    'BackgroundLightCyan': "\033[106m",
    'BackgroundWhite': "\033[40m"
}


def chop_microseconds(delta: datetime.timedelta) -> datetime.timedelta:
    return delta - datetime.timedelta(microseconds=delta.microseconds)


def get_timer_delta(start_timer: float, with_ms: bool = False) -> datetime.timedelta:
    """
    :param start_timer: begin time
    :param with_ms: if microseconds neened - set to true. else: lowest is seconds
    :return:
    time = get_timer()
    print('Total run time {}'.format(get_timer_delta(start_timer=time)))
    "Total run time 0:00:00.000026"
    """
    end_timer = get_timer()
    d = datetime.timedelta(seconds=(end_timer - start_timer))
    if not with_ms:
        d = chop_microseconds(d)
    return d


def get_timer() -> float:
    return timer()


def timer_action(seconds: int, action: str = '', tabs: int = 1) -> None:
    """ counts till seconds or block"""
    if seconds is None:
        input('{}Press "Enter" key for {}...'.format(tabs * '\t', action))
    else:
        time_in_future = datetime.datetime.now() + datetime.timedelta(seconds=seconds)
        print('{}{} IN: {}'.format(tabs * '\t', action, seconds), end='', flush=True)
        while time_in_future > datetime.datetime.now():
            time.sleep(1)
            seconds -= 1
            print(' {}'.format(seconds), end='', flush=True)
        print('')
    return


def get_current_date_hour() -> str:
    """
    running start time - good for long programs
    :return:
    """
    now = datetime.datetime.now()
    current_time = now.strftime('%d-%m-%Y %H:%M:%S')
    return current_time


def get_pc_name() -> str:
    """
    :return: pc name as str
    """
    import platform
    return platform.uname()[1]


def get_mac_address(add_semi_colons: bool = True) -> str:
    """
    :return: pc mac as str
    print(get_mac_address(add_semi_colons=True)) # 70:4D:7B:8A:65:EE
    print(get_mac_address(add_semi_colons=False)) # 704D7B8A65EE
    """
    from uuid import getnode as get_mac
    mac = get_mac()
    mac = "{:X}".format(mac)  # turn to Hex
    if add_semi_colons:
        mac = ':'.join(mac[i:i + 2] for i in range(0, 12, 2))
    return mac


def get_cuda_version() -> str:
    """
    :return: cuda version if found on environment variables
    """
    if 'CUDA_PATH' in os.environ:
        cuda_v = os.path.basename(os.environ['CUDA_PATH'])
    else:
        cuda_v = 'No CUDA_PATH found'
    return cuda_v


def make_cuda_invisible() -> None:
    """
        disable(the -1) gpu 0
        TODO support hiding multiple gpus
    """
    os.environ['CUDA_VISIBLE_DEVICES'] = '-1, 0'
    return


def start_profiler() -> cProfile.Profile:
    """
    starts profiling
    :return: profiling object that is needed for end_profiler()
    """
    pr = cProfile.Profile()
    pr.enable()
    return pr


def end_profiler(pr: cProfile.Profile, rows: int = 10) -> str:
    """
    profiling output
    :param pr: object returned from start_profiler()
    :param rows: how many rows to print sorted by 'cumulative' run time
    :return:
    """
    pr.disable()
    s = io.StringIO()
    ps = pstats.Stats(pr, stream=s).sort_stats('cumulative')
    ps.print_stats(rows)
    return s.getvalue()


def set_seed(seed: int = 42) -> None:
    """
    :param seed: setting numpy and random seeds
    :return:
    """
    np.random.seed(seed)
    random.seed(seed)
    return


def get_tensorflow_version() -> str:
    import tensorflow as tf
    string = '* TensorFlow Version {}'.format(tf.__version__)
    return string


def main_wrapper(
        main_function: Callable,
        cuda_off: bool = False,
        torch_v: bool = False,
        tf_v: bool = False,
        cv2_v: bool = False,
        with_profiler: bool = False
) -> None:
    """
    :param main_function: the function to run
    :param cuda_off: make gpu invisible and force run on cpu
    :param torch_v: print torch version
    :param tf_v: print tensorflow version
    :param cv2_v: print opencv version
    :param with_profiler: run profiler
    :return:
    """
    print(LINES)
    start_timer = get_timer()

    # make_cuda_invisible()
    print('main_wrapper:')
    print('* Run started at {}'.format(get_current_date_hour()))
    print('* Python Version {}'.format(sys.version))
    print('* Working Dir: {}'.format(os.getcwd()))
    print('* Computer Name: {}'.format(get_pc_name()))
    print('* Computer Mac: {}'.format(get_mac_address()))
    cuda_msg = '* CUDA Version: {}'.format(get_cuda_version())
    if cuda_off:
        make_cuda_invisible()
        cuda_msg += ' (Turned off)'
    print(cuda_msg)

    if torch_v:
        try:
            from wizzi_utils.torch_tools import get_torch_version
            print(get_torch_version())
        except (ImportError, ModuleNotFoundError, NameError) as err:
            string = '* {}'.format(err)
            print(string)
    if tf_v:
        try:
            print(get_tensorflow_version())
        except (ImportError, ModuleNotFoundError, NameError) as err:
            string = '* {}'.format(err)
            print(string)
    if cv2_v:
        try:
            from wizzi_utils.open_cv_tools import get_cv_version
            print(get_cv_version())
        except (ImportError, ModuleNotFoundError, NameError) as err:
            string = '* {}'.format(err)
            print(string)

    print('Function {} started:'.format(main_function))
    print(LINES)
    pr = start_profiler() if with_profiler else None
    main_function()
    if with_profiler:
        print(end_profiler(pr))
    print(LINES)
    print('Total run time {}'.format(get_timer_delta(start_timer)))
    return


def add_data(var, data_chars: int, float_precision: int = 0) -> str:
    """
    Aux function for to_str()
    :param var:
    :param data_chars:
    :param float_precision:
    :return:
    """

    if isinstance(var, (float, int)):
        if isinstance(var, float) and float_precision > 0:
            data_str_raw = '{:,.' + str(float_precision) + 'f}'
            data_str_raw = data_str_raw.format(var)
        else:
            data_str_raw = '{:,}'.format(var)
    else:
        data_str_raw = str(var)
    data_str_raw = data_str_raw.replace('\n', '').replace('  ', '')
    if data_chars == 0:  # all data
        data_chars = len(data_str_raw) + 1
    elif data_chars < 0:  # no data
        data_chars = 0
    # else keep positive chars

    data_str = ': {}'.format(data_str_raw[:data_chars])
    if len(data_str_raw) > data_chars > 0:
        data_str += ' ...too long'
    return data_str


def to_str(var, title: str, data_chars: int = 100, float_precision: int = 2, recursive: bool = False) -> str:
    """
    :param var: the variable
    :param title: the title (usually variable name)
    :param data_chars: how many char to print.
        -1: none
         0: all
        +0: maximum 'data_chars' (e.g. data_chars=50 and |str(var)|=100 - first 50 chars)
    :param float_precision: round number if possible(float, np array...). 0 for no rounding
    :param recursive: round number if possible(float, np array...). 0 for no rounding

    examples:
        print(to_str(var=32223123123123123, title='my_int'))
        print(to_str(var=3.2, title='my_float'))
        print(to_str(var=1234567890.223123123123123123, title='my_long_float', float_precision=3))
        print(to_str(var='a', title='my_str'))
        print(to_str(var=[], title='my_empty_list'))
        print(to_str(var=[112312312, 3, 4], title='1d list of ints'))
        print(to_str(var=[1, 3123123], title='1d list of ints no data', data_chars=-1))  # no data
        print(to_str(var=[11235] * 1000, title='1d long list'))
        print(to_str(var=(1239, 3, 9), title='1d tuple'))
        print(to_str(var=[[1231.2123123, 15.9], [3.0, 7.55]], title='2d list'))
        print(to_str(var=[(1231.2123123, 15.9), (3.0, 7.55)], title='2d list of tuples'))
        b = np.array([[1231.123122, 15.9], [3.0, 7.55]])
        print(to_str(var=b, title='2d np array'))
        cv_img = np.zeros(shape=[480, 640, 3], dtype=np.uint8)
        print(to_str(var=cv_img, title='cv_img'))
        print(to_str(var={'a': [1213, 2]}, title='dict of lists'))
        print(to_str(var={'a': [{'k': [1, 2]}, {'c': [7, 2]}]}, title='nested dict'))
    :return: informative string of the variable
    """
    type_s = str(type(var)).replace('<class \'', '').replace('\'>', '')  # clean type name
    string = '{}({})'.format(title, type_s)  # base: title and type

    if isinstance(var, (int, float, str)):
        if hasattr(var, "__len__"):
            string = string.replace(')', ',len={})'.format(var.__len__()))
        string += add_data(var, data_chars, float_precision)

    elif isinstance(var, (list, tuple)):
        string = string.replace(')', ',len={})'.format(var.__len__()))
        string += add_data(var, data_chars)
        if recursive and len(var) > 0:  # recursive call
            string += '\n\t{}'.format(
                to_str(var=var[0], title='{}[0]'.format(title), data_chars=data_chars, float_precision=float_precision,
                       recursive=recursive))

    elif isinstance(var, np.ndarray):
        string = string.replace(')', ',shape={},dtype={})'.format(var.shape, var.dtype))
        if var.dtype == np.float64 and float_precision > 0:
            string += add_data(np.round(var, float_precision).tolist(), data_chars, float_precision=0)
        else:
            string += add_data(var.tolist(), data_chars, float_precision=0)
        if recursive and len(var) > 0:  # recursive call
            string += '\n\t{}'.format(
                to_str(var=var[0], title='{}[0]'.format(title), data_chars=data_chars, float_precision=float_precision,
                       recursive=recursive))

    elif isinstance(var, dict):
        string = string.replace(')', ',len={},keys={})'.format(var.__len__(), list(var.keys())))
        string += add_data(var, data_chars)
        if recursive and len(var) > 0:  # recursive call
            first_key = next(iter(var))
            string += '\n\t{}'.format(
                to_str(var=var[first_key], title='{}[{}]'.format(title, first_key), data_chars=data_chars,
                       float_precision=float_precision, recursive=recursive))

    else:  # all unidentified elements get default print (title(type): data)
        string += add_data(var, data_chars)
    return string


def save_np(t: np.array, path: str, ack_print: bool = True, tabs: int = 0):
    """
    to save a dict - see np.savez(path, name1=a, name2=b ... ) - loading the same
    :param t: numpy array
    :param path: suffix '.npy' added automatically
    :param ack_print:
    :param tabs:
    :return:
    """
    np.save(path, t)
    if ack_print:
        print('{}Saved to: {}'.format(tabs * '\t', path))
    return


def load_np(path: str, ack_print: bool = True, tabs: int = 0) -> np.array:
    """
    :param path:
    :param ack_print:
    :param tabs:
    :return: numpy array
    """
    t = np.load(path)
    if ack_print:
        print('{}Loaded: {}'.format(tabs * '\t', path))
    return t


def save_pkl(path: str, data_dict: dict, ack: bool = True, tabs: int = 0) -> None:
    file_obj = open(path, "wb")
    pickle.dump(data_dict, file_obj)
    file_obj.close()
    if ack:
        print('{}Saved to: {}'.format(tabs * '\t', path))
    return


def load_pkl(path: str, ack: bool = True, tabs: int = 0) -> dict:
    file_obj = open(path, "rb")
    data_dict = pickle.load(file_obj)
    file_obj.close()
    if ack:
        print('{}Loaded: {}'.format(tabs * '\t', path))
    return data_dict


def get_uniform_dist_by_dim(A):
    lows = np.min(A, axis=0)
    highs = np.max(A, axis=0)
    return lows, highs


def get_normal_dist_by_dim(A):
    means = np.mean(A, axis=0)
    stds = np.std(A, axis=0)
    return means, stds


def np_normal(shape: tuple, mius: list, stds: list) -> np.array:
    """
    e.g.
        d = 2
        A = np.zeros((3, d))
        A[0][0], A[0][1] = 0, 1000
        A[1][0], A[1][1] = 1, 2000
        A[2][0], A[2][1] = 5, 4000
        means, stds = get_normal_dist_by_dim(A)
        print(means, stds)
        A_2 = np_normal(shape=(500, d), mius=means, stds=stds)
        print(A_2.shape)
        print(get_normal_dist_by_dim(A_2))
    """
    ret = np.random.normal(loc=mius, scale=stds, size=shape)
    return ret


def np_uniform(shape: tuple, lows: [list, int], highs: [list, int]) -> np.array:
    """
    e.g.
        d=2
        A = np.zeros((3, d))
        A[0][0], A[0][1] = 0, 1000
        A[1][0], A[1][1] = 1, 2000
        A[2][0], A[2][1] = 5, 4000
        lows, highs = get_uniform_dist_by_dim(A)
        print(lows, highs)
        A_2 = np_uniform(shape=(10, d), lows=lows, highs=highs)
        print(A_2.shape)
        print(get_normal_dist_by_dim(A_2))
    """
    ret = np.random.uniform(low=lows, high=highs, size=shape)
    return ret


def np_random_integers(low: int, high: int, size: tuple) -> np.array:
    ret = np.random.random_integers(low=low, high=high, size=size)
    return ret


def augment_x_y_numpy(X: np.array, y: np.array) -> np.array:
    """ creates A=X|y """
    assert X.shape[0] == y.shape[0], 'row count must be the same'
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.reshape(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.reshape(y.shape[0], 1)
    A = np.column_stack((X, y))
    return A


def de_augment_numpy(A: np.array) -> (np.array, np.array):
    """ creates X|y=A """
    if len(A.shape) == 1:  # A is 1 point. change from size (n) to size (1,n)
        A = A.reshape(1, A.shape[0])
    X, y = A[:, :-1], A[:, -1]
    if len(X.shape) == 1:  # change x size()=[n,] to size()=[n,1]
        X = X.reshape(X.shape[0], 1)
    if len(y.shape) == 1:  # change y size()=[n,] to size()=[n,1]
        y = y.reshape(y.shape[0], 1)
    return X, y


def nCk(n: int, k: int, as_int: bool = False):
    """
    n choose k
    :return: if as_int True: the result of nCk, else the combinations of nCk

    e.g.
    A = np.random.randint(low=-10, high=10, size=(3, 2))
    print('A={}'.format(A.tolist()))

    # let's iterate on every 2 different indices of A
    combs_count = utils.nCk(len(A), k=2, as_int=True)
    print('|A| choose 2={}:'.format(combs_count)) # result is 3

    combs_list = utils.nCk(len(A), k=2)  # result is [[0, 1], [0, 2], [1, 2]]
    for i, comb in enumerate(combs_list):
        print('\tcomb {}={}. A[comb]={}'.format(i, comb, A[comb].tolist()))
    """
    range_list = np.arange(0, n, 1)
    combs = list(combinations(range_list, k))
    combs = [list(comb) for comb in combs]
    if as_int:
        combs = len(combs)
    return combs


def redirect_std_start():
    """
    redirect all prints to summary_str
    e.g.
        old_stdout, summary_str = redirect_std_start()
        print('blablabla')
        string = redirect_std_finish(old_stdout, summary_str)
        print('captured output: {}'.format(string))
    :return:
    """
    old_stdout = sys.stdout
    sys.stdout = summary_str = io.StringIO()
    return old_stdout, summary_str


def redirect_std_finish(old_stdout, summary_str) -> str:
    """ redirect all prints back to std out and return a string of what was captured"""
    sys.stdout = old_stdout
    return summary_str.getvalue()


def get_line_number(depth: int = 1) -> str:
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.lineno)
    except IndexError:
        pass
    return ret_val


def get_function_name(depth: int = 1) -> str:
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.function)
    except IndexError:
        pass
    return ret_val


def get_file_name(depth: int = 1) -> str:
    ret_val = ''
    try:
        scope_1_back = inspect.stack()[depth]  # stack()[0] is this function
        ret_val = '{}'.format(scope_1_back.filename)
    except IndexError:
        pass
    return ret_val


def get_function_name_and_line(depth: int = 1) -> str:
    # +1 because of this function
    ret_val = '{} line {}'.format(get_function_name(depth + 1), get_line_number(depth + 1))
    return ret_val


def add_color(string: str, color: str = 'Red') -> str:
    """
    :param string:
    :param color: from color map
    # todo add option for combination (like bold, underlined ...)
    # todo change to lower case
    # todo color numbers maybe wrong
    e.g.
        my_str = 'hello colorful world!'
        print(my_str)
        print(add_color(my_str, color='Red'))
        print(add_color(my_str, color='Blue'))
        print(add_color(my_str, color='Bold'))
    to see all colors:
        for k, v in CONST_COLOR_MAP.items():
            print('{}{}{}'.format(v, k, CONST_COLOR_MAP['ResetAll']))
    """

    if color in CONST_COLOR_MAP:
        # concat (color, string, reset tag)
        string = '{}{}{}'.format(CONST_COLOR_MAP[color], string, CONST_COLOR_MAP['ResetAll'])
    return string


def init_logger(logger_path: str = None):
    """
    example:
        init_logger('./log.txt')
        log_print('line 1')
        flush_logger()
        log_print('line 2')
        log_print('line 3')
        close_logger()
    """
    global logger
    logger = open(file=logger_path, mode='w', encoding='utf-8') if logger_path is not None else None
    print(type(logger))
    return logger


def flush_logger() -> None:
    """ good for loops - writes every iteration if used """
    global logger
    if logger is not None:
        logger.flush()
    return


def log_print(string: str) -> None:
    global logger
    print(string)
    if logger is not None:
        logger.write('{}\n'.format(string))
    return


def log_print_dict(my_dict) -> None:
    for k, v in my_dict.items():
        log_print('\t{}: {}'.format(k, v))
    return


def close_logger() -> None:
    global logger
    if logger is not None:
        logger.close()
    return


def create_dir_if_not_exists(data_root: str, ack: bool = True, tabs: int = 1):
    if os.path.exists(data_root):
        if ack:
            print('{}{} exists'.format(tabs * '\t', data_root))
    else:
        os.mkdir(data_root)
        if ack:
            print('{}{} created'.format(tabs * '\t', data_root))
    return


def sleep(seconds: int):
    time.sleep(seconds)
    return


def reverse_tuple_or_list(orig: [tuple, list]) -> [tuple, list]:
    """
    :param orig: list or tuple
    :return: dst: reversed list or tuple
    e.g.
    my_tuple = (0, 0, 255)
    print(my_tuple)
    print(reverse_tuple_or_list(my_tuple))
    my_list = [0, 0, 255]
    print(my_list)
    print(reverse_tuple_or_list(my_list))
    """
    dst = orig[::-1]
    return dst


def get_time_stamp(format_s: str) -> str:
    """
    :param format_s:
    examples:
    "%Y_%m_%d_%H_%M_%S_%f" >>> 2020_07_19_13_36_24_597247
    "%Y_%m_%d" >>> 2021_04_07
    :return:
    """
    date_time_obj = datetime.datetime.now()
    timestamp_str = date_time_obj.strftime(format_s)
    return timestamp_str


def main():
    print(to_str(var=32223123123123123, title='my_int'))
    print(to_str(var=3.2, title='my_float'))
    print(to_str(var=3.2123123, title='my_float', float_precision=4))
    print(to_str(var=1234567890.223123123123123123, title='my_long_float', float_precision=3))
    print(to_str(var='a', title='my_str'))
    print(to_str(var=[], title='my_empty_list'))
    print(to_str(var=[112312312, 3, 4], title='1d list of ints', recursive=True))
    print(to_str(var=[1, 3123123], title='1d list of ints no data', data_chars=-1, recursive=True))  # no data
    print(to_str(var=[1.0000012323, 3123123.22454875123123], title='1d list of ints no data', float_precision=7,
                 recursive=True))
    print(to_str(var=np.array([1.0000012323, 3123123.22454875123123], dtype=float), title='1d list of ints no data',
                 float_precision=7, recursive=True))
    print(to_str(var=[11235] * 1000, title='1d long list', recursive=True))
    print(to_str(var=(1239, 3, 9), title='1d tuple', recursive=True))
    print(to_str(var=[[1231.2123123, 15.9], [3.0, 7.55]], title='2d list', recursive=True))
    print(to_str(var=[(1231.2123123, 15.9), (3.0, 7.55)], title='2d list of tuples', recursive=True))
    b = np.array([[1231.123122, 15.9], [3.0, 7.55]])
    print(to_str(var=b, title='2d np array', recursive=True))
    cv_img = np.zeros(shape=[480, 640, 3], dtype=np.uint8)
    print(to_str(var=cv_img, title='cv_img', recursive=True))
    print(to_str(var={'a': [1213, 2]}, title='dict of lists', recursive=True))
    print(to_str(var={'a': [{'k': [1, 2]}, {'c': [7, 2]}]}, title='nested dict', recursive=True))
    return


if __name__ == '__main__':
    main_wrapper(
        main_function=main,
        cuda_off=True,
        torch_v=True,
        tf_v=False,
        cv2_v=True,
        with_profiler=False
    )



try:
    from wizzi_utils.temp_sub_package import temp_module
    print('init temp_sub_package worked')
except ModuleNotFoundError as e:
    print('Missing some modules')
    print('Error: {}'.format(e))


def temp_func_test():
    print(temp_module.temp_func())
    return


def test_all():
    temp_func_test()
    return



import numpy as np


def select_coreset_frequency(
        P: np.array,
        SP: np.array,
        coreset_size: int,
        W_in: np.array = None,
) -> (np.array, np.array):
    """
    P, SP, W_in - all the same type: numpy
    :param P:  your points. if you have labels: from the shape X|y
    :param SP: your points sensitivity.
    :param coreset_size: the size of desired coreset
    :param W_in: weights of P. if not wighted input - ones(n)
    :return:
    """
    n = P.shape[0]
    assert coreset_size <= n
    if W_in is None:
        W_in = np.ones(n)

    prob = SP / SP.sum()
    indices_set = set()
    indices_list = []

    cnt = 0
    while len(indices_set) < coreset_size:
        i = np.random.choice(a=n, size=1, p=prob).tolist()[0]
        indices_list.append(i)
        indices_set.add(i)
        cnt += 1

    hist = np.histogram(indices_list, bins=range(n + 1))[0].flatten()
    indices_list = np.nonzero(hist)[0]
    C = P[indices_list, :]

    W_out = (W_in[indices_list].T * hist[indices_list]).T
    W_out = (W_out.T / (prob[indices_list] * cnt)).T
    return C, W_out


def select_coreset(A: np.array, SP: np.array, coreset_size: int) -> (np.array, np.array):
    """
    :param A: nx(d+1) points. can be torch\numpy
    :param SP: numpy array of nx1 sensitivities for each p in A
    :param coreset_size: size of the coreset
    :return:
    C: subset of A of size coreset_size x (d+1)
    W: weights of each c in C
    """
    n = len(A)
    sp_sum = SP.sum()
    sp_normed = SP / sp_sum
    # replace: False -> unique values, True  -> with reps
    C_ind = np.random.choice(a=range(0, n), size=coreset_size, p=sp_normed, replace=True)
    C = A[C_ind]
    W = sp_sum / (coreset_size * SP[C_ind])
    return C, W


def main():
    return


if __name__ == '__main__':
    main()



try:
    from wizzi_utils.temp_sub_package.temp_module import *
    print('init temp_sub_package worked')
except ModuleNotFoundError as e:
    print('Missing some modules')
    print('Error: {}'.format(e))




import socket
import numpy as np


def open_server(server_address: tuple = ('localhost', 10000), ack: bool = True, tabs: int = 1) -> socket:
    if ack:
        print('{}Opening server on IP,PORT {}'.format(tabs * '\t', server_address))
    sock = socket.socket(socket.AF_INET, socket.SOCK_STREAM)
    sock.bind(server_address)
    sock.listen(1)
    return sock


def get_host_name(ack: bool = True, tabs: int = 1) -> str:
    """ :return: hostname """
    hostname = socket.gethostname()
    if ack:
        print("{}Computer Name: {}".format(tabs * '\t', hostname))
    return hostname


def get_ipv4(ack: bool = True, tabs: int = 1) -> str:
    """ :return ipv4 address of this computer """
    ipv4 = socket.gethostbyname(get_host_name(ack=True, tabs=0))
    if ack:
        print("{}Computer IP Address: {}".format(tabs * '\t', ipv4))
    return ipv4


def send_msg(connection: socket, buflen: int, data: str, msg_end: str) -> None:
    data_e = str.encode(data + msg_end)
    data_e_len = len(data_e)
    for i in range(0, data_e_len, buflen):
        chunk_i = data_e[i:i + buflen]
        connection.send(chunk_i)
    return


def receive_msg(connection: socket, buflen: int, msg_end: str) -> (str, int):
    data_in = ''
    saw_end_delimiter = False
    while not saw_end_delimiter:
        data_in += connection.recv(buflen).decode('utf-8')
        if not data_in:
            break  # empty transmission
        if data_in.endswith(msg_end):
            data_in = data_in.replace('$#$#', '')
            saw_end_delimiter = True

    # data_in = connection.recv(buflen).decode('utf-8')
    data_in_len = len(data_in)
    return data_in, data_in_len


def buffer_to_str(data: str, prefix: str, tabs: int = 1, max_chars: int = 100) -> str:
    """
    :param data:
    :param prefix:
    :param tabs:
    :param max_chars:
    :return:
    data = 'hi server, how you doing???'  # len(data)==27
    print(buffer_to_str(data, prefix='client1', tabs=0, max_chars=27))
    print(buffer_to_str(data, prefix='client1', tabs=0, max_chars=26))
    print(buffer_to_str(data, prefix='client1', tabs=1, max_chars=15))
    """
    data_len = len(data)
    data_chars = data_len + 1 if data_len <= max_chars else max_chars

    msg = '{}{}: {} (bytes sent={})'.format(tabs * '\t', prefix, data[:data_chars], data_len)
    if data_len > max_chars:
        msg += ' ... message is too long'
    return msg


def rounds_summary(times: list) -> None:
    """
    :param times:
    :return:
    times_ti = []
    for t in range(10):
        print('t={}:'.format(t))
        begin_timer_i = misc_tools.get_timer()
        # do_work of round t
        misc_tools.sleep(seconds=0.03)
        end_timer_i = misc_tools.get_timer()
        times_ti.append(end_timer_i - begin_timer_i)
        print('\tDONE round={}: total time={}'.format(t, misc_tools.get_timer_delta(begin_timer_i)))
    rounds_summary(times_ti)
    output:
    Rounds Summary:
        Total rounds = 10
        Total run time = 0.328 seconds
        Avg   run time = 0.033 seconds (std = 0.004)
        30.50 FPS
    """
    if len(times) > 0:  # try print time avg before exit
        print('\nRounds Summary:')
        print('\tTotal rounds = {}'.format(len(times)))
        print('\tTotal run time = {:.3f} seconds'.format(np.sum(times)))
        print('\tAvg   run time = {:.3f} seconds (std = {:.3f})'.format(np.mean(times), np.std(times)))
        if np.mean(times) > 0.0001:
            print('\t{:.2f} FPS'.format(1 / np.mean(times)))
    return


def main():
    return


if __name__ == '__main__':
    main()



from distutils.core import setup

setup(
    name='wizzi_utils',  # How you named your package folder (MyLib)
    packages=[
        'wizzi_utils',
        'wizzi_utils/temp_sub_package',
        'wizzi_utils/temp_sub_package/test'
    ],  # main package and sub packages
    version='4.4.1',  # Start with a small number and increase it with every change you make
    license='MIT',  # Chose a license from here: https://help.github.com/articles/licensing-a-repository
    description='some handy tools',  # Give a short description about your library
    author='Gilad Eini',  # Type in your name
    author_email='giladEini@gmail.com',  # Type in your E-Mail
    url='https://github.com/2easy4wizzi/2021wizzi_utils',  # Provide either the link to your github or to your website
    # TODO update on new release
    download_url='https://github.com/2easy4wizzi/2021wizzi_utils/archive/refs/tags/v_4.4.1.tar.gz',
    keywords=[  # Keywords that define your package best
        'debug tools',
        'cuda',
        'torch',
        'cv2',
        'tensorflow'
    ],
    install_requires=[  # TODO add pip installed libraries
        'datetime',
        'typing',
        'numpy',
    ],
    classifiers=[
        'Development Status :: 3 - Alpha',
        # Chose either "3 - Alpha", "4 - Beta" or "5 - Production/Stable" as the current state of your package
        'Intended Audience :: Developers',  # Define that your audience are developers
        'Topic :: Software Development :: Build Tools',
        'License :: OSI Approved :: MIT License',  # Again, pick a license
        'Programming Language :: Python :: 3.6'
    ],
)



"""
package wizzi utils:
    pip install wizzi_utils
    of
    pip install wizzi_utils --upgrade

to import all functions:
    import wizzi_utils as wu
    # the above will give you all functions and tests
    # everything in misc_tools must work
    # everything else, e.g. torch_tools, will work only if you have all
        the modules installed.

examples:
    # direct access to a module - must work
    print(wu.misc_tools.to_str(var=1, title='my_int'))

    # access to a function in the main module misc_tools - must work
    print(wu.to_str(var=2, title='my_int'))

    # access to a function in the torch module - will work if you have torch
        and the rest of the modules needed installed
    print(wu.tt.to_str(var=3, title='my_int'))

    # access to a function in the matplotlib module - same as torch above
    print(wu.pyplt.get_RGB_color(color_str='r'))
"""

__version__ = 'X.Y.Z'

from wizzi_utils.misc_tools import *  # misc tools

# from .misc_tools import *  # misc tools
# __all__ = ['misc_tools']
# import .torch_tools as tt

from wizzi_utils import temp_sub_package as temp_pac

try:
    from wizzi_utils import torch_tools as tt  # torch tools
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import pyplot_tools as pyplt  # pyplot tools
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import algorithms as algs  # known algorithms
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import open_cv_tools as cvt  # cv2 tools
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import coreset_tools as cot  # coreset tools
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import json_tools as jt  # coreset tools
except ModuleNotFoundError:
    pass

try:
    from wizzi_utils import socket_tools as st  # coreset tools
except ModuleNotFoundError:
    pass



import os
import json


def json_to_string(j: dict, indent: int = -1, sort_keys: bool = False) -> str:
    """
    :param j: dict
    :param indent: how many indents
    :param sort_keys: sort dict keys
    :return: string rep of j
    e.g.
    cfg = {'x': 3, 'a': 'dx'}
    print('one liner:')
    print(json_to_string(cfg, indent=-1, sort_keys=False))  # 1 liner
    print('pretty print - 4 indentations:')
    print(json_to_string(cfg, indent=4, sort_keys=False))  # pretty print - 4 indentations
    print('pretty print - 0 indentations:')
    print(json_to_string(cfg, indent=0, sort_keys=False))  # pretty print - 0 indentations
    print('one liner sorted keys:')
    print(json_to_string(cfg, indent=-1, sort_keys=True))  # 1 liner
    """
    if indent == -1:
        indent = None
    return json.dumps(j, indent=indent, sort_keys=sort_keys)


def string_to_json(j_str: str) -> json:
    """
    e.g.
    j_str = '{"x": 3, "a": "dx"}'
    j = string_to_json(j_str)
    print(type(j))
    """
    return json.loads(j_str)


def load_json(file_path: str, ack: bool = True) -> dict:
    """ loads a dict in json format from path """
    ret_dict = {}
    if os.path.exists(file_path):
        ret_dict = json.load(open(file_path))
        if ack:
            print('{} loaded (type {})'.format(file_path, type(ret_dict)))
            print(json_to_string(ret_dict, indent=4))
    else:
        print('file {} doesnt exists'.format(file_path))
    return ret_dict


def load_jsons(files_path: list, ack: bool = True) -> dict:
    """
    loads several of json files format from paths and concat to one dict
    asserts if a key found on 2 of the files
    """
    all_in_one_dict = {}
    len_keys_json = 0
    for file_path in files_path:
        j = load_json(file_path, ack=False)
        len_keys_json += len(j)
        all_in_one_dict.update(j)
    if ack:
        print('{} loaded (type {})'.format(files_path, type(all_in_one_dict)))
        print(json_to_string(all_in_one_dict, indent=4))
    assert len_keys_json == len(all_in_one_dict), 'Duplicated keys found: {}'.format(files_path)
    return all_in_one_dict


def save_json(file_path: str, j: dict, indent: int = -1, sort_keys: bool = False, ack: bool = True) -> None:
    json.dump(j, open(file_path, 'w'), indent=indent, sort_keys=sort_keys)
    if ack:
        print('{} saved successfully'.format(file_path))
        print(json_to_string(j, indent=indent, sort_keys=sort_keys))
    return


def get_key_by_value(d: dict, value) -> str:
    """
    :param d: dict
    :param value:
    Notice that it will return the first key of the value given. if value not unique...
    :return: the key of the value
    e.g.
    j = {"x": 3, "a": "dx"}
    print(get_key_by_value(j, value="dx"))
    print(get_key_by_value(j, value=3))
    """
    key = None
    for k, v in d.items():
        if v == value:
            key = k
            break
    return key


def main():
    return


if __name__ == '__main__':
    main()
